#include "merlin_hw.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <core/decoders/merlin.h>
#include <core/x_core.h>
#include <core/macros/mouse.h>

static inline void timeout(void)
{
	x_keyboard_data.n_keys = 0;
	x_keyboard_data.mod_byte = 0;
	x_keyboard_data.keyboard_changed = true;
	x_ir_reader.timeout = 0;
	X_MOUSE_RESET();
	x_ir_reader.ignore_keys_until_next_timeout = false;
	x_ir_reader.log = false;
}

x_ir_reader_t x_ir_reader =
{
	.signal = false,
	.log = false,
	.data = {0},
	.data_pos = 0,
	.keyboard_changed = false,
	.timeout = 0,
	.ignore_keys_until_next_timeout = false,
};

#if (ARCH == ARCH_XMEGA)

#include <usb/xmega/Q/osc/osc.h>
#include <usb/LUFA/Drivers/USB/USB.h>
#include <usb/LUFA/Platform/Platform.h>
#include <usb/LUFA/Drivers/USB/Class/Device/HIDClassDevice.h>
#include <usb/xmega/Descriptors.h>
#include <usb/xmega/usb.h>

static inline void usb_init(void)
{
	int2x_ppl_USB48(((F_CPU/100000) + 10)/20); //Takt automatisch nach F_CPU setzen
	USB_Init();
	
	CDC_Device_CreateStream(&VirtualSerial_CDC_Interface, stdout);
}

static inline void usb_loop(void)
{
	CDC_Device_ReceiveByte(&VirtualSerial_CDC_Interface);
	
	CDC_Device_USBTask(&VirtualSerial_CDC_Interface);
	HID_Device_USBTask(&Mouse_HID_Interface);
	HID_Device_USBTask(&Keyboard_HID_Interface);
	USB_USBTask();
}


#elif (ARCH == ARCH_AVR8)

#include <avr/wdt.h>
#include <usb/v-usb/usbdrv.h>
#include <usb/mega/usb.h>
#include <avr/io.h>
#include <util/delay.h>

#include <avr/iomx8.h>
#include <core/scancodes.h>

static inline void usb_loop(void)
{
	uint8_t i;
	wdt_reset();
	usbPoll();
	if(usbInterruptIsReady())
	{
		if(x_mouse_info.mouse_move_x || x_mouse_info.mouse_move_y || x_mouse_info.buttons_pressed || x_mouse_info.mouse_wheel)
		{
			mouse_report.report_id = 0x4d;
			mouse_report.axis_x = x_mouse_info.mouse_move_x;
			mouse_report.axis_y = x_mouse_info.mouse_move_y;
			mouse_report.wheel = x_mouse_info.mouse_wheel;
			mouse_report.buttons = x_mouse_info.buttons_pressed;
			x_mouse_info.mouse_move_y = 0;
			x_mouse_info.mouse_move_x = 0;
			x_mouse_info.mouse_wheel = 0;
			usbSetInterrupt((void *)&mouse_report, sizeof(mouse_report));
			return;
		}
		keyboard_report.report_id = 0x02;
		for(i = 0; i < x_keyboard_data.n_keys; i++)
		{
			//Alle gerade gedrueckten Keys
			keyboard_report.keycode[i] = x_keyboard_data.keys[i];
		}
		for( ; i < MERLIN_MAX_KEYS; i++)
		{
			keyboard_report.keycode[i] = 0;
		}
		//sowie aktive Modifiers
		keyboard_report.modifier = x_keyboard_data.mod_byte;
		
		usbSetInterrupt((void *)&keyboard_report, sizeof(keyboard_report));
	}
}

static inline void usb_init(void)
{
	wdt_enable(WDTO_1S);
	/* Even if you don't use the watchdog, turn it off here. On newer devices,
	 * the status of the watchdog (on/off, period) is PRESERVED OVER RESET!
	 */
	/* RESET status: all port bits are inputs without pull-up.
	 * That's the way we need D+ and D-. Therefore we don't need any
	 * additional hardware initialization.
	 */
	uint8_t i;
	usbInit();
	usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
	
	i = 0;
	while(--i)
	{             /* fake USB disconnect for > 250 ms */
		wdt_reset();
		_delay_ms(1);
	}
	usbDeviceConnect();
}

#else
#error "ERROR: Arch not supported"
#endif

void merlin_global_init(void)
{
	
	MERLIN_LEDS_SETUP();
	
	

	usb_init();
	
	_delay_ms(1);
	
	MERLIN_TIMER_SETUP();
	// 	MERLIN_ENABLE_TIMER_IRQ();
	
	MERLIN_EXT_IRQ_SETUP();
	MERLIN_EXT_IRQ_ENABLE();
	
	MERLIN_ENABLE_IRQ_GLOBAL();
	
	x_core_init();
}


void merlin_main_loop(void)
{
	static uint16_t no_signal_count = 0;
	
	static uint8_t x_codes[MERLIN_MAX_KEYS];
	
	static uint8_t length;
	int8_t tmp;
	
	//Kein signal und messung aktiv
	if(x_ir_reader.log && !x_ir_reader.signal)
	{
		no_signal_count++; //ein bisschen luft lassen
		if(no_signal_count > SIGNAL_TIMEOUT_ITERATIONS)
		{
			MERLIN_LED2_TGL();

			MERLIN_EXT_IRQ_DISABLE();
			if(x_ir_reader.data_pos > 5 && !x_ir_reader.ignore_keys_until_next_timeout) //Ignore incomplete codes
			{
				tmp = decode(x_codes);
				if(tmp != -1)
				{
					length = tmp;
					translate(x_codes, length);
				}
			}

			x_ir_reader.log = false;
			x_ir_reader.data_pos = 0;
			x_ir_reader.timeout = 0;
			MERLIN_EXT_IRQ_ENABLE();
			no_signal_count = 0;
		}
	}
	else
	{
		//Timeout-Handling, damit falls das Stopsignal nicht empfangen wird, der Empfaenger nicht immer
		//weiter Daten an den Host sendet.
		x_ir_reader.timeout++;
		if(x_ir_reader.timeout >= TIMEOUT_ITERATIONS)
		{
			MERLIN_LED1_TGL();
			timeout();
		}
	}
	if(x_ir_reader.signal)
	{
		no_signal_count = 0;
	}
	x_ir_reader.signal = false;
	
	usb_loop();
#if (ARCH == ARCH_AVR8)
	_delay_us(3);
#endif
}

//TODO: ein timer wäre schöner aber v-usb mag das bisher leider nicht

ISR(MERLIN_EXT_IRQ_VECT)
{
	uint16_t res;
	MERLIN_TIMER_GETRES(res);
	if(x_ir_reader.data_pos < MERLIN_MAX_DATA_LENGTH)
	{
		x_ir_reader.data[x_ir_reader.data_pos] = res;
		x_ir_reader.data_pos++;
	}
	x_ir_reader.signal = true;  //Has clock this perriode
	x_ir_reader.log = true;     //Currently reading
	x_ir_reader.timeout = 0;    //TODO: timeout gibt es bisher nicht wenn es spikes gibt...
}